// Copyright 2013 Yangqing Jia

#include <stdint.h>
#include <leveldb/db.h>
#include <pthread.h>

#include <string>
#include <vector>

#include "caffe/layer.hpp"
#include "caffe/util/io.hpp"
#include "caffe/vision_layers.hpp"

using std::string;

namespace caffe {

template <typename Dtype>
HMCDataLayer<Dtype>::~HMCDataLayer<Dtype>() { }

template <typename Dtype>
void HMCDataLayer<Dtype>::SetUp(const vector<Blob<Dtype>*>& bottom,
      vector<Blob<Dtype>*>* top) {
	CHECK_EQ(bottom.size(), 2) << "HMC Data Layer takes two input blobs.";
	CHECK_EQ(top->size(), 2) << "Data Layer takes two blobs as output.";


	if (this->layer_param_.has_meanfile() == 0 && this->layer_param_.cropsize() == 1) {
		LOG(FATAL) << "Current implementation requires meanfile and cropsize to be "
			<< "set at the same time.";
	}


	// data
	(*top)[0]->Reshape(bottom[0]->num(), bottom[0]->channels(), bottom[0]->height(), bottom[0]->width());
	// label
	(*top)[1]->Reshape(bottom[1]->num(), bottom[1]->channels(), bottom[1]->height(), bottom[1]->width());

	// check if we want to have mean
	if (this->layer_param_.has_meanfile()) {
		BlobProto blob_proto;
		LOG(INFO) << "Loading mean file from" << this->layer_param_.meanfile();
		ReadProtoFromBinaryFile(this->layer_param_.meanfile().c_str(), &blob_proto);
		data_mean_.FromProto(blob_proto);
		CHECK_EQ(data_mean_.num(), 1);
		CHECK_EQ(data_mean_.channels(), bottom[0]->channels());
		if (this->layer_param_.cropsize()) {
			CHECK_EQ(this->layer_param_.cropsize(), bottom[0]->height()) << "Crop size should be equal to sample data height.";
			CHECK_EQ(this->layer_param_.cropsize(), bottom[0]->width()) << "Crop size should be equal to sample data width.";
		} else {
			CHECK_EQ(data_mean_.height(), bottom[0]->height());
			CHECK_EQ(data_mean_.width(), bottom[0]->width());
		}
	} else {
		// Simply initialize an all-empty mean.
		data_mean_.Reshape(1, bottom[0]->channels(), bottom[0]->height(), bottom[0]->width());
		Dtype* mean = this->data_mean_.mutable_cpu_data();
		for (int i = 0; i < data_mean_.count(); i++) {
			mean[i] = 0.0;
		}
	}
}

template <typename Dtype>
void HMCDataLayer<Dtype>::Forward_cpu(const vector<Blob<Dtype>*>& bottom,
      vector<Blob<Dtype>*>* top) {
	Dtype* top_data = (*top)[0]->mutable_cpu_data();
	Dtype* top_label = (*top)[1]->mutable_cpu_data();
	const Dtype* bottom_data = bottom[0]->cpu_data();
	const Dtype* bottom_label = bottom[1]->cpu_data();
	const Dtype scale = this->layer_param_.scale();
	const Dtype alpha = this->layer_param_.alpha();
	const int num = bottom[0]->num();
	const int size = bottom[0]->count()/bottom[0]->num();
	const int cropsize = this->layer_param_.cropsize();
	const int height = bottom[0]->height();
	const int width = bottom[0]->width();
	const int channels = bottom[0]->channels();
	const Dtype* mean = this->data_mean_.cpu_data();

	if (cropsize) {
		int h_off = (height - cropsize) / 2;
		int w_off = (width - cropsize) / 2;
		for (int itemid = 0; itemid < num; ++itemid) {
			for (int c = 0; c < channels; ++c) {
				for (int h = 0; h < cropsize; ++h) {
					for (int w = 0; w < cropsize; ++w) {
						top_data[((itemid * channels + c) * cropsize + h) * cropsize + w]
						         = (bottom_data[((itemid * channels + c) * cropsize + h) * cropsize + w]
						           - mean[(c * height + h + h_off) * width + w + w_off]) * scale;
					}
				}
			}
			top_label[itemid] = bottom_label[itemid];
		}

	} else {

		for (int itemid = 0; itemid < num; ++itemid) {
			for (int j = 0; j < size; ++j) {
				top_data[itemid * size + j] =
				(bottom_data[itemid * size + j] - mean[j]) * scale + alpha;
			}
			top_label[itemid] = bottom_label[itemid];
		}
	}
}

// The backward operations.
template <typename Dtype>
Dtype HMCDataLayer<Dtype>::Backward_cpu(const vector<Blob<Dtype>*>& top,
      const bool propagate_down, vector<Blob<Dtype>*>* bottom) {
	const Dtype* top_diff = top[0]->cpu_diff();
	const Dtype* top_data = top[0]->cpu_data();
	Dtype* bottom_diff = (*bottom)[0]->mutable_cpu_diff();
	const Dtype scale = this->layer_param_.scale();
	for (int i = 0; i < top[0]->count(); i++) {
		bottom_diff[i] = top_diff[i]*scale;
	}
	return Dtype(0.);
}

template <typename Dtype>
void HMCDataLayer<Dtype>::Forward_gpu(const vector<Blob<Dtype>*>& bottom,
      vector<Blob<Dtype>*>* top) {
	Forward_cpu(bottom, top);
}

// The backward operations.
template <typename Dtype>
Dtype HMCDataLayer<Dtype>::Backward_gpu(const vector<Blob<Dtype>*>& top,
      const bool propagate_down, vector<Blob<Dtype>*>* bottom) {
	return Backward_cpu(top, propagate_down, bottom);
}

INSTANTIATE_CLASS(HMCDataLayer);

}  // namespace caffe
