// Copyright 2013 Yangqing Jia

#include <algorithm>
#include <cfloat>
#include <vector>

#include "caffe/layer.hpp"
#include "caffe/vision_layers.hpp"
#include "caffe/util/math_functions.hpp"

using std::max;

namespace caffe {

template <typename Dtype>
void HMCLossLayer<Dtype>::SetUp(const vector<Blob<Dtype>*>& bottom,
      vector<Blob<Dtype>*>* top) {
	CHECK_EQ(bottom.size(), 2) << "HMC Loss Layer takes two blobs as input.";
	CHECK_EQ(top->size(), 1) << "HMC Loss Layer takes a single blob as output.";

    (*top)[0]->Reshape(bottom[0]->num(), 1, 1, 1);
};

template <typename Dtype>
void HMCLossLayer<Dtype>::Forward_cpu(const vector<Blob<Dtype>*>& bottom,
    vector<Blob<Dtype>*>* top) {
	const Dtype* bottom_data = bottom[0]->cpu_data();
	const Dtype* label = bottom[1]->cpu_data();
	Dtype* top_output = (*top)[0]->mutable_cpu_data();
	int num = bottom[0]->num();
	int dim = bottom[0]->count() / bottom[0]->num();
	
	//int test_id = 0;
	for (int i = 0; i < num; ++i) {
		top_output[i] = -bottom_data[i*dim + static_cast<int>(label[i])];
		//if(top_output[i] != 0)
		//	test_id = 1;
	}
	//LOG(INFO) << "TEST_ID is "  << test_id;
}

template <typename Dtype>
void HMCLossLayer<Dtype>::Forward_gpu(const vector<Blob<Dtype>*>& bottom,
    vector<Blob<Dtype>*>* top) {
	// TODO(Yangqing): implement the GPU version of softmax.
	return Forward_cpu(bottom, top);
}

template <typename Dtype>
Dtype HMCLossLayer<Dtype>::Backward_cpu(const vector<Blob<Dtype>*>& top,
    const bool propagate_down,
    vector<Blob<Dtype>*>* bottom) {
	// First, compute the diff
	Dtype* bottom_diff = (*bottom)[0]->mutable_cpu_diff();
	const Dtype* label = (*bottom)[1]->cpu_data();
	const Dtype* top_output = top[0]->cpu_data();
	int num = (*bottom)[0]->num();
	int dim = (*bottom)[0]->count() / num;
	
	for (int i = 0; i < (*bottom)[0]->count(); ++i) {
		bottom_diff[i] = 0.0;
	}

	Dtype loss = 0.;
	for (int i = 0; i < num; ++i) {
		bottom_diff[i * dim + static_cast<int>(label[i])] -= 1.0;
		loss += top_output[i];
	}

	return loss/num;
}

template <typename Dtype>
Dtype HMCLossLayer<Dtype>::Backward_gpu(const vector<Blob<Dtype>*>& top,
    const bool propagate_down, vector<Blob<Dtype>*>* bottom) {
  // TODO(Yangqing): implement the GPU version of softmax.
  return Backward_cpu(top, propagate_down, bottom);
}

INSTANTIATE_CLASS(HMCLossLayer);


}  // namespace caffe
